#!/usr/bin/env python
#   Copyright 2012 Gary Wright http://identi.ca/wrightsolutions
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import with_statement
from fabric.api import abort, cd, run, settings
from fabric.contrib.console import confirm
from fabric.contrib.files import exists

""" res.succeeded is preferred for newer fabric, but not res.failed
is used where backward compatibility is preferable. """

cmd_wget = "wget -nd -m -l 0 "
cmd_gpg_recv = "gpg --keyserver keys.gnupg.net --recv-keys "
cmd_gpg_verify = "gpg --verify --no-tty --batch "
cmd_gpg_list = "gpg --list-key --no-tty --batch "
dir_fab = "primerange-benchmark.fab"
dirpath_fab = "~/" + dir_fab


def tilde_expand(path):
    if path.startswith('~/'):
        path_expanded = run("dirname ~/.")
    else:
        return path
    return path_expanded


def rmdir_existing(dirpath,ignore_missing=True):
    if dirpath.startswith('~/'):
        """ set function default return to True """
        rmdir_flag = True
    else:
        """ As a safety feature against dangerous rm -fR
        commands, any directory path that does not begin
        with ~/ will be rejected (function returns False) """
        return False
    exists_dirpath = True
    with settings(warn_only=True):
        """ Tilde expansion only applies when '~' is unquoted """
        exists_dirpath = exists(tilde_expand(dirpath))
        if not exists_dirpath:
            return ignore_missing
    if exists_dirpath:
        run("rm -vfR " + dirpath)
    return True


def abort_with_cleanup(dirpath,abort_message,ignore_missing=True):
    rmdir_existing(dirpath,ignore_missing)
    abort(abort_message)
    return


def mkdir_existing(dirpath,remove_existing=True):
    if dirpath.startswith('~/'):
        """ set function default return to True """
        mkdir_flag = True
    else:
        """ As a safety feature against dangerous rm -fR
        commands, any directory path that does not begin
        with ~/ will be rejected (function returns False) """
        return False
    exists_dirpath = True
    with settings(warn_only=True):
        """ Tilde expansion only applies when '~' is unquoted """
        # exists_dirpath = exists(dirpath,verbose=True)
        exists_dirpath = exists(tilde_expand(dirpath))
        #print "exists_dirpath=%s, remove_existing=%s\n" \
        #% (exists_dirpath,remove_existing)
    if exists_dirpath and remove_existing:
        run("rm -vfR " + dirpath)
    mkdir_res = run("mkdir " + dirpath)
    if mkdir_res.failed:
        mkdir_flag = False
    return mkdir_flag


def gpg_recv_and_verify(signed_file):
    gpgres = run(cmd_gpg_list + "CE0222D0EB8CFF0E")
    if gpgres.failed:
       gpgres = run(cmd_gpg_recv + "CE0222D0EB8CFF0E")
       if not gpgres.failed:
       	  abort("Aborting as preparation for gpg verification failed.")
    gpgres = run(cmd_gpg_verify + signed_file + ".sig")	 
    #gpgres = contains(tar,'02 Feb 2012 18:48:46 GMT using DSA key ID EB8C')	 
    return gpgres


def populate_input_file(file_source, file_destination, rows=9999999):
    populate_flag = False
    """ As a safety feature against dangerous cp or output
    redirect commands, any file_destination that does not begin
    with ~/ or /tmp/ will be rejected (function returns False) """
    if file_destination.startswith('~/'):
        populate_flag = True
    elif file_destination.startswith('/tmp/'):
        populate_flag = True

    if (rows > 23537):
        res = run("cp -p " + file_source + " " + file_destination)
    else:
        res = run("head -n " + str(rows) + " " + file_source + \
                      " > " + file_destination)
    return populate_flag


def primerange_get_and_build(rows='9999999'):
    """ If you want fabric command line interaction with an argument
    then expect it to be a string. May need to int() what you get """
    rows_limit = int(rows)
    #print "primerange_get_and_build rows=%d\n" % rows_limit
    tar = 'primerange-partial__20111008prerelease.tar.gz'
    tar_url = "http://download-mirror.savannah.gnu.org/releases/primerange/%s.sig" % tar
    mkdir_existing(dirpath_fab)
    with settings(warn_only=True):
	 """
	 res = local('curl -I http://www.wrightsolutions.co.uk/',
                     capture=True)
	 """
         with cd(dirpath_fab):
             res = run(cmd_wget + tar_url)
             if not res.failed:
	 	 tar_url = "http://download-mirror.savannah.gnu.org/releases/primerange/%s" % tar
		 res = run(cmd_wget + tar_url)
		 """ capture=True can only be used for local - not okay for run() """
		 if not res.failed:
		    	res = gpg_recv_and_verify(tar)
	
    if res.failed and not confirm("Download of primerange failed. Continue anyway?"):
        abort_with_cleanup(dirpath_fab,"Aborting at user request.")

    run("which gcc pv")

    if (rows_limit > 23537):
        print "Timed benchmark run being prepared.\n"
    else:
        print "Timed run being prepared with input rows=%d.\n" % rows_limit

    with cd(dirpath_fab):
        res = run("tar -zxf " + tar)
        mpztwodir = "primerange/mpz_two_large_factors_of_known_form"
        pop_res = populate_input_file(mpztwodir + "/*23538*.composite2largefac", \
                                "/tmp/primerange.composite2largefac", rows_limit)
        if pop_res:
            mpztwo = "mpz_two_large_factors_of_known_form-" + \
                "integers_to_factor_are_uniform_sizeEven2468THIRTYFOURMILLlimit"
            res = run("cp -p " + mpztwodir + "/" + mpztwo + ".c . ")
            res = run("gcc " + mpztwo + ".c -o " + mpztwo + ".o -lm -lgmp")
        else:
            abort_with_cleanup(dirpath_fab,"Aborting as problem populating the input file.")            

        timed_run_flag = False
        if (rows_limit < 1000):
            timed_run_flag = True
        elif (rows_limit < 2500):
            if confirm("Timed run may take an hour to complete. Continue?"):
                timed_run_flag = True
        elif confirm("Timed run may take between an hour and 15 hours to complete. Continue?"):
            timed_run_flag = True

        if timed_run_flag:
            res = run("date;time ./" + mpztwo + ".o | pv -f -i 60 --buffer-size 1k -b > /dev/null;date" )
        """ fgrep of output will perhaps not show immediately due to buffering / other delay """
        #res = run("date;time ./" + mpztwo + ".o | fgrep 'processing begun...';date" )

