#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2013 Gary Wright wrightsolutions.co.uk/contact
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Tested against fabric 1.4.3
# http://ftp.uk.debian.org/debian/pool/main/f/fabric/fabric_1.4.3-1.debian.tar.gz

from __future__ import with_statement
from fabric.api import abort, cd, env, get, hide, put, run, settings, sudo
from fabric.contrib.console import confirm
from fabric.contrib.files import exists,append
from os import path
import re
from string import ascii_lowercase,digits,printable,punctuation,whitespace
try:
    from . import cmdbin
except ValueError:
    import cmdbin

""" res.succeeded is preferred for newer fabric, but not res.failed
is used where backward compatibility is preferable. """

CMD_APT = "/usr/bin/apt-get "
CMD_APT_NORECOMMENDS = "/usr/bin/apt-get --no-install-recommends "
CMD_APT_ASSUMEYES = "/usr/bin/apt-get --assume-yes "
CMD_APT_ASSUMEYES_NORECOMMENDS = "/usr/bin/apt-get --no-install-recommends -y "
CMD_APT_ASSUMEYES_NORECOMMENDS_CONFOLD = "/usr/bin/apt-get \
 -o DPkg::Options::=--force-confold --no-install-recommends -y "
CMD_APT_ASSUMEYES_NORECOMMENDS_CONFMISS = "/usr/bin/apt-get \
 -o Dpkg::Options::=--force-confmiss --reinstall --no-install-recommends -y "
CMD_APT_UPDATE = "/usr/bin/apt-get update "
CMD_DPKG = "/usr/bin/dpkg "
CMD_DPKG_SELECTIONS = "/usr/bin/dpkg --get-selections "
CMD_DPKG_STATUS = "/usr/bin/dpkg -s "
CMD_DPKG_SEARCH = "/usr/bin/dpkg --search "
CMD_DPKG_QUERY_STATUS = "/usr/bin/dpkg-query -W --showformat='${Status;-30}'"
CMD_DPKG_QUERY_STATUS3 = "/usr/bin/dpkg-query -W \
--showformat='${Status;-30} ${Package;-30} ${Version;-30}' "
CMD_DPKG_QUERY_STATUS4 = "/usr/bin/dpkg-query -W \
--showformat='${Installed-Size;-10}\t${Status;-30}\t${Section;-15}\t${Package;-30}\n' "
CMD_NETSELECT_APT = "/usr/bin/netselect-apt"

RE_STATUS_INSTALLED = re.compile('Status:.*installed')
RE_NO_PACKAGES_MATCHING = re.compile(r'no\spackages\sfound\smatching\b')
RE_OK_INSTALLED = re.compile(r'\bok\sinstalled\b')
SELECTIONS2LIST = ['install','hold']
""" Above has 2 entries. Below has 4 as also includes 
status that indicate purged or marked for deinstall """
SELECTIONS4LIST = ['install','hold','deinstall','purge']

_list_dot_d_array = []
_sources_list_object = None


def list_dot_d_array_extend(extend_with):
    """ Used to as a helper to maintain a list of sources .list files
    to support cleanup functions which might want to undo any sources .list
    additions on abort / other controlled situation
    Example: files_directories.py abort_with_cleanup()
    """
    global _list_dot_d_array
    extended = True
    try:
        if hasattr(extend_with, "__iter__") and type(extend_with) == type(''):
            extend_with = [extend_with]
            # Could have coded up list_dot_d_array_append(append_string) but
            # just fix up any api misuse by making list from string instead.
        _list_dot_d_array.extend(extend_with)
    except Exception,e:
        extended = False
    return extended


def get_list_dot_d_array():
    return _list_dot_d_array


def has_backports():
    has_flag = False
    with settings(warn_only=True):
        REGEX = '.*(/sources.list|/sources.list.d/.*\.list)'
        find_line = "{0} /etc/apt/ {1} {2}{3}{2} -exec cat {4}".format(cmdbin.CMD_FIND,
                                                                       cmdbin.REGEX_OPTS_EGREP,
                                                                       chr(39),REGEX,cmdbin.EXEC_TAIL)
        ros_res = cmdbin.run_or_sudo(find_line)
        res_lines = cmdbin.run_res_lines_stripped(ros_res)
        for line in res_lines:
            if line.startswith('deb '):
                if '-backports ' in line:
                    #print "MATCH: {0}".format(line)
                    has_flag = True
        """ find /etc/apt/ -regextype posix-egrep -regex \
        '.*(/sources.list|/sources.list.d/.*\.list)' -exec cat {} \;
        Above is a longhand representation of our find_line construction.
        What follows is okay but will bork when nothing in sources.list.d
        grep_line = "{0} '^deb ' {1}/sources.list {1}/sources.list.d/*.list".format(
            cmdbin.CMD_EGREP,'/etc/apt') """
    return has_flag


def dpkg_status_lines_list(packagename=None,status='installed'):
    """ The status 'installed' is assumed and is the only status currently
    supported. This function could be enhanced to accept other values
    as second argument by using addition re similar to RE_STATUS_INSTALLED
    """
    if not packagename:
        return []

    with settings(warn_only=True):
        """ cmdbin.run_or_sudo("{0} {1}".format(CMD_DPKG_STATUS,packagename)) """
        status_res = run("{0} {1}".format(CMD_DPKG_STATUS,packagename))
        #cmdbin.print_success_and_failed(status_res)

    status_list = []
    if status_res.succeeded:
        status_list = [ ln for ln in status_res.split('\n') \
                            if RE_STATUS_INSTALLED.match(ln) ]

    return status_list


def dpkg_status_installed(packages_list=[],return_type='flag'):
    """ By default this function returns True or False,
    but giving second argument of 'outstanding' you would
    instead have a list returned.
    """
    packages_list_count = len(packages_list)
    if packages_list_count < 1:
        return False
    installed_list_count = 0
    installed_res = True
    outstanding_list = []
    for package in packages_list:
        if set(package).issubset(cmdbin.SET_PRINTABLE):
            res_lines = []
            with settings(warn_only=True):
                with hide('warnings'):
                    status_res = run("{0} {1}".format(CMD_DPKG_QUERY_STATUS4,package))
                    res_lines = cmdbin.run_res_lines_stripped(status_res)
            if status_res.succeeded:
                for line in res_lines:
                    line_search = line[:42]
                    if RE_NO_PACKAGES_MATCHING.search(line_search):
                        outstanding_list.append(package)
                    elif RE_OK_INSTALLED.search(line_search):
                        installed_list_count += 1
                    else:
                        outstanding_list.append(package)
            else:
                outstanding_list.append(package)
        else:
            installed_res = False
            break
    #print len(outstanding_list)
    if installed_res and installed_list_count == packages_list_count:
        outstanding_list = []
    else:
        installed_res = False
    if return_type == 'outstanding':
        return outstanding_list
    return installed_res


def dpkg_status_installed1(packagename):
    """ Return a simple True or False - does the package
    appear to be ok installed?
    """
    if packagename is None or len(packagename) < 2:
        return False
    installed = False
    query_line = "{0} {1}".format(CMD_DPKG_QUERY_STATUS4,packagename)
    with settings(warn_only=True):
        """ Warning only now set. Next hide """
        with hide('stdout','warnings'):
            status_res = run("{0} {1}".format(CMD_DPKG_QUERY_STATUS4,packagename))
            for line in cmdbin.run_res_lines_stripped_searching(status_res,' installed'):
                installed = True
                break
    return installed


def dpkg_installed_version(packages_list=[],return_type=None):
    """ By default this function returns True or False,
    but giving second argument of 'outstanding' you would
    instead have a list returned.
    """
    version_list = []
    packages_list_count = len(packages_list)
    if packages_list_count < 1:
        return version_list
    installed_list_count = 0
    installed_res = True
    outstanding_list = []
    for package in packages_list:
        if set(package).issubset(cmdbin.SET_PRINTABLE):
            res_lines = []
            with settings(warn_only=True):
            #with hide('output'):
                status_res = run("{0} {1}".format(CMD_DPKG_QUERY_STATUS3,package))
                res_lines = cmdbin.run_res_lines_stripped(status_res)
            if status_res.succeeded:
                for line in res_lines:
                    line_search = line[:31]
                    if RE_NO_PACKAGES_MATCHING.search(line_search):
                        outstanding_list.append(package)
                    elif RE_OK_INSTALLED.search(line_search):
                        line_array = line.split(package)
                        if len(line_array) > 1:
                            installed_list_count += 1
                            version_list.append(line_array[1].strip())
                    else:
                        outstanding_list.append(package)
            else:
                outstanding_list.append(package)
        else:
            installed_res = False
            break
    #print len(outstanding_list)
    if installed_res and installed_list_count == packages_list_count:
        outstanding_list = []
        return version_list
    else:
        installed_res = False
    if return_type == 'outstanding':
        return outstanding_list
    return version_list


def selection_lines_including(lines,including=None,excluding=None,colon_discard=True):
    """ Including has a higher priority than excluding.
    If including is given special keyword 'anything' then
    given value for excluding is ignored
    """
    return_lines = []
    if including == 'anything':
        # Keyword 'anything' means all 4 status will be included 
        including = SELECTIONS4LIST
        excluding = None
    elif including is None and excluding is None:
        # None for 'excluding' means all 4 status unless 'including' is meaningful
        including = SELECTIONS4LIST
        excluding = None
    if including is not None:
        if type(including) == type(''):
            including = [including]
        if len(set(including).difference(SELECTIONS4LIST)) > 0:
            raise Exception("'including' list should be limited to {0}".format(SELECTIONS4LIST))
    if excluding is not None:
        if type(excluding) == type(''):
            excluding = [excluding]
        if len(set(excluding).difference(SELECTIONS4LIST)) > 0:
            raise Exception("'including' list should be limited to {0}".format(SELECTIONS4LIST))
    if colon_discard:
        for line in lines:
            line_array = line.strip().split()
            if len(line_array) > 0:
                status = line_array[-1]
                joined = ''.join(line_array[:-1]).strip()
                if including is not None and status in including:
                    return_lines.append(joined.split(chr(58))[0])
                elif excluding is not None and status not in excluding:
                    return_lines.append(joined.split(chr(58))[0])
            else:
                # Design choice - be forgiving about blank lines
                pass
    else:
        for line in lines:
            line_array = line.strip().split()
            if len(line_array) > 0:
                status = line_array[-1]
                joined = ''.join(line_array[:-1]).strip()
                if including is not None and status in including:
                    return_lines.append(joined)
                elif excluding is not None and status not in excluding:
                    return_lines.append(joined)
            else:
                # Design choice - be forgiving about blank lines
                pass
    return return_lines


def selections_including(including_status=None,colon_discard=True):
    """ chr(58) is colon (:)
    colon_discard=True says not interested in :i386 markers (and similar)
    that sometimes follow package name to indicate arch of i386/amd64/whatever
    Discarding would mean 'e2fslibs:i386' would be made to appear in returned
    list as simply 'e2fslibs'
    """
    if including_status is None:
        # Assume want 'install' and 'hold' unless instructed otherwise
        including_status = SELECTIONS2LIST
    elif including_status == 'anything':
        # Keyword 'anything' means all 4 status will be included 
        including_status = SELECTIONS4LIST
    elif type(including_status) == type(''):
        including_status = [including_status]
    res_lines = []
    with hide('output'):
        selections_res = run(CMD_DPKG_SELECTIONS)
        res_lines = cmdbin.run_res_lines_stripped(selections_res)
    including_lines = []
    if colon_discard:
        for line in res_lines:
            line_array = line.strip().split()
            status = line_array[-1]
            joined = ''.join(line_array[:-1]).strip()
            if status in including_status:
                including_lines.append(joined.split(chr(58))[0])

    else:
        for line in res_lines:
            line_array = line.strip().split()
            status = line_array[-1]
            joined = ''.join(line_array[:-1]).strip()
            if status in including_status:
                including_lines.append(joined)
    return including_lines


def selections_excluding(excluding_status=None,colon_discard=True):
    """ chr(58) is colon (:)
    colon_discard=True says not interested in :i386 markers (and similar)
    that sometimes follow package name to indicate arch of i386/amd64/whatever
    Discarding would mean 'e2fslibs:i386' would be made to appear in returned
    list as simply 'e2fslibs'
    """
    if excluding_status is None:
        # Assume want 'install' and 'hold' unless instructed otherwise
        excluding_status = ['deinstall','purge']
    elif excluding_status == 'anything':
        # Keyword 'anything' means all 4 status will be included 
        return []
        #excluding_status = SELECTIONS4LIST
    elif type(excluding_status) == type(''):
        excluding_status = [excluding_status]
    selections_res = run(CMD_DPKG_SELECTIONS)
    res_lines = []
    with hide('output'):
        selections_res = run(CMD_DPKG_SELECTIONS)
        res_lines = cmdbin.run_res_lines_stripped(selections_res)
    excluding_lines = []
    if colon_discard:
        for line in res_lines:
            line_array = line.strip().split()
            status = line_array[-1]
            joined = ''.join(line_array[:-1]).strip()
            if status in excluding_status:
                pass
            else:
                excluding_lines.append(joined.split(chr(58))[0])

    else:
        for line in res_lines:
            line_array = line.strip().split()
            status = line_array[-1]
            joined = ''.join(line_array[:-1]).strip()
            if status in excluding_status:
                pass
            else:
                excluding_lines.append(joined)
    return excluding_lines


def packages_summary(print_flag=False,verbose=False):
    return


def packages_marked_hold_summary(print_flag=False,verbose=False):
    #selections_including(['hold'],True):
    selections = selections_including('hold',True)
    selections_count = len(selections)
    if print_flag:
        if selections_count > 0:
            if verbose == True:
                print "Packages held:"
                for pkg in selections:
                    print "- {0}".format(pkg)
            else:
                print "Packages held: {0}".format(selections_count)
        else:
            print "Packages held: None. No packages are marked 'hold'"
    return selections


def held_packages_summary(print_flag=False,verbose=False):
    """ Using held rather than hold in the function name can
    lead to asking for packages 'held' rather than an
    include_list of ['hold']. Use the fuller function name
    (shown on next line) directly if you want to be very clear.
    """
    return packages_marked_hold_summary(print_flag,verbose)


def source_list_lines_debian(pretty_dict,repolabel,backports_main_only=True):
    source_list_lines = None
    if repolabel == 'backports' and backports_main_only:
        source_list_lines = """
# added by fabutil
deb http://ftp.debian.org/{id} {release}-backports main
deb-src http://ftp.debian.org/{id} {release}-backports main
""".format(**pretty_dict)
    elif repolabel == 'backports':
        source_list_lines = """
# added by fabutil
deb http://ftp.debian.org/{id} {release}-backports main contrib non-free
deb-src http://ftp.debian.org/{id} {release}-backports main contrib non-free
""".format(**pretty_dict)
    elif repolabel in ['contrib','non-free']:
        source_list_lines = """
# added by fabutil
deb http://http.debian.net/{id}/ {release} {repolabel}
deb-src http://http.debian.net/{id}/ {release} {repolabel}
""".format(**pretty_dict)
    else:
        pass
    return source_list_lines


def sources_list_lines(list_dot_d):
    sl_lines = None
    (pname, pretty_name_id, pretty_name_word) = cmdbin.pretty_name_tuple()
    if pretty_name_id == 'debian':
        """ See www.debian.org/doc/debian-policy/ch-archive.html for more about
        distinction between main and other repositories
        """
        if list_dot_d in ['backports','contrib','non-free']:
            pretty_dict = {'id': pretty_name_id,
                           'release': pretty_name_word,
                           'repolabel': list_dot_d}
            sl_lines = source_list_lines_debian(pretty_dict,list_dot_d)
        else:
            pass
    elif pretty_name_id == 'ubuntu':
        """ Unlike Debian which distinguishes repos by 'freedom to modify/distribute'
        Ubuntu distinguishes repos by level of support by Canonical company.
        A firm grasp on this distinction will help you avoid simplistic maps, but next
        is one to get you started: (contrib+non-free)->company filter->(multiverse+restricted)
        """
        if list_dot_d in ['universe','multiverse']:
            pretty_dict = {'id': pretty_name_id,
                           'release': pretty_name_word,
                           'verse': list_dot_d}
            sl_lines = """
# added by fabutil
deb http://archive.ubuntu.com/{id}/ {release} {verse}
deb-src http://archive.ubuntu.com/{id}/ {release} {verse}
""".format(**pretty_dict)
    else:
        pass
    return sl_lines


def missing_conf_via_reinstall(filepath):
    if filepath is None:
        return False
    if chr(47) in filepath:
        """ Regular file including some path element (/)
        Here forward slash is hardcoded. This software is intended for remote
        systems that are GNU/Linux or GNU variants. """
        pass
    else:
        return False
    reinstalled = False
    ros_res = run("{0}{1}".format(CMD_DPKG_SEARCH,filepath))
    packagename = None
    for line in cmdbin.run_res_lines_stripped_searching(ros_res,chr(47)):
        first = line.split()[0]
        if len(first) > 0:
            packagename = first.rstrip(chr(58))
            break
    installed = False
    if packagename is not None:
        installed = dpkg_status_installed1(packagename)
    if installed:
        apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS_CONFMISS,
                                            packagename)
        ros_res = cmdbin.run_or_sudo(apt_line)
        if len(cmdbin.run_res_lines_stripped_searching(ros_res,'etting')) > 0:
            reinstalled = True
    return reinstalled


def missing_conf_via_reinstall_middle(filepath,middle='local'):
    destination = cmdbin.filepath_suffixed_isocompact(filepath,middle)
    copy_line = "{0} {1} {2};".format(cmdbin.CMD_CP_UPDATE_NOCROSS_PRESERVETIME,
                                      filepath,destination)
    rm_line = "{0} {1};".format(cmdbin.CMD_RM_BIGI,filepath)
    with settings(warn_only=True):
        """ Warning only now set. Next hide """
        with hide('stdout','warnings'):
            copy_res = cmdbin.run_or_sudo(copy_line)
            if copy_res.failed:
                return False
            rmres = cmdbin.run_or_sudo(rm_line)
            if rmres.failed:
                return False
    return missing_conf_via_reinstall(filepath)


def aptget_line(packages_list=[],assumeyes=False,
                recommends=True,confold=False,precheck=False):

    """ Returning an empty string indicates nothing to do (no packages need installing)
    For exceptions/unexpected situations then raise Exceptions
    """

    packages_list_count = len(packages_list)
    if packages_list_count < 1:
        raise Exception('Unexpected value as first argument')

    if precheck:
        packages_list = dpkg_status_installed(packages_list,'outstanding')
        if len(packages_list) < 1:
            return ''

    packages = ' '.join(packages_list)
    apt_line = "{0} install {1}".format(CMD_APT,packages)
    if assumeyes:
        if recommends is False:
            if confold:
                apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS_CONFOLD,
                                                    packages)
            else:
                apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS,
                                                    packages)
        else:
            apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES,packages)
    else:
        if recommends is False:
            apt_line = "{0} install {1}".format(CMD_APT_NORECOMMENDS,packages)
    return apt_line


def requires_packages(packages_list=[],assumeyes=False,
                      recommends=True,confold=False):
    """ Installs required packages.
    Unlike the 'one by one' variant requires_dpkg_package() this function
    has no list_dot_d argument. So if you did need to install from
    'contrib', 'non-free', 'universe', 'multiverse' and you want automatic
    repository adding, then instead use requires_dpkg_package()
    """
    requires_res = False
    ros_line = aptget_line(packages_list,assumeyes,recommends,confold,True)
    ros_res = cmdbin.run_or_sudo(ros_line)
    if ros_res.succeeded:
        requires_res = True  
    return requires_res


def requires_dpkg_package(packagename=None,assumeyes=False,
                          recommends=True,confold=False,list_dot_d=None):
    if packagename:
        requires_res = True
    else:
        return False

    if len(dpkg_status_lines_list(packagename)) > 0:
        return True

    requires_res = False

    with settings(warn_only=True):
        apt_line = "{0} install {1}".format(CMD_APT,packagename)
        if assumeyes:
            if recommends is False:
                if confold:
                    apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS_CONFOLD,
                                                        packagename)
                else:
                    apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS,
                                                        packagename)
            else:
                apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES,packagename)
        else:
            if recommends is False:
                apt_line = "{0} install {1}".format(CMD_APT_NORECOMMENDS,packagename)
        ros_res = cmdbin.run_or_sudo(apt_line)
        if ros_res.succeeded:
            requires_res = True  
        elif list_dot_d:
            sl_filename = "{0}.list".format(list_dot_d)
            sl_lines = None
            (pname, pretty_name_id, pretty_name_word) = cmdbin.pretty_name_tuple()
            if pretty_name_id == 'debian':
                """ See www.debian.org/doc/debian-policy/ch-archive.html for more about
                distinction between main and other repositories
                """
                if list_dot_d in ['contrib','non-free']:
                    pretty_dict = {'id': pretty_name_id,
                                   'release': pretty_name_word,
                                   'repolabel': list_dot_d}
                    sl_lines = """
# added by fabutil
deb http://http.debian.net/{id}/ {release} {repolabel}
deb-src http://http.debian.net/{id}/ {release} {repolabel}
""".format(**pretty_dict)
            elif pretty_name_id == 'ubuntu':
                """ Unlike Debian which distinguishes repos by 'freedom to modify/distribute'
                Ubuntu distinguishes repos by level of support by Canonical company.
                A firm grasp on this distinction will help you avoid simplistic maps, but next
                is one to get you started: (contrib+non-free)->company filter->(multiverse+restricted)
                """
                if list_dot_d in ['universe','multiverse']:
                    pretty_dict = {'id': pretty_name_id,
                                   'release': pretty_name_word,
                                   'verse': list_dot_d}
                    sl_lines = """
# added by fabutil
deb http://archive.ubuntu.com/{id}/ {release} {verse}
deb-src http://archive.ubuntu.com/{id}/ {release} {verse}
""".format(**pretty_dict)
            else:
                pass
            if sl_lines:
                put_flag = cmdbin.put_using_named_temp_sudo("/etc/apt/sources.list.d/{0}".format(sl_filename),
                                                       sl_lines)
                if put_flag:
                    list_dot_d_array_extend(sl_filename)

                    cmdbin.run_or_sudo(CMD_APT_UPDATE)

                    apt_line = "{0} install {1}".format(CMD_APT,packagename)
                    if assumeyes:
                        if recommends is False:
                            if confold:
                                apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS_CONFOLD,
                                                                    packagename)
                            else:
                                apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS,
                                                                    packagename)
                        else:
                            apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES,packagename)
                    else:
                        if recommends is False:
                            apt_line = "{0} install {1}".format(CMD_APT_NORECOMMENDS,packagename)
                    ros_res = cmdbin.run_or_sudo(apt_line)
                    if ros_res.succeeded:
                        requires_res = True
                else:
                    requires_res = False
                    """ Above else: for clarity rather than necessity """

    status_list = dpkg_status_lines_list(packagename)
    print "{0} - {1}".format(packagename, status_list)
    # if len(dpkg_status_lines_list(packagename)) > 0:
    if len(status_list) > 0:
        return True

    return requires_res


def requires_backported_package(packagename=None,assumeyes=False,
                                recommends=True,confold=False):
    requires_res = False
    if packagename is not None:
        if len(packagename) > 0 and type(packagename) == type(''):
            # chr(61) is equals
            if chr(61) in packagename:
                """ Unlike requires_backported_packages() this
                function does not support package=version style
                requests
                """
                pass
            else:
                requires_res = True

    if requires_res is False:
        return False
    requires_res = True

    if len(dpkg_status_lines_list(packagename)) > 0:
        return True

    BACKPORTS='backports'

    has_backports_flag = False
    if has_backports():
        has_backports_flag = True
    else:
        (pname, pretty_name_id, pretty_name_word) = cmdbin.pretty_name_tuple()
        pretty_dict = {'id': pretty_name_id,
                       'release': pretty_name_word,
                       'repolabel': BACKPORTS}
        sl_filename = "{0}.list".format(BACKPORTS)
        sl_lines = source_list_lines_debian(pretty_dict,BACKPORTS)
        if sl_lines:
            sl_filepath = "/etc/apt/sources.list.d/{0}".format(sl_filename)
            put_flag = cmdbin.put_using_named_temp_sudo(sl_filepath,sl_lines)
            if put_flag:
                has_backports_flag = True
                list_dot_d_array_extend([sl_filename])
                #print get_list_dot_d_array()
                with hide('output'):
                    ros_res = cmdbin.run_or_sudo(CMD_APT_UPDATE)
                    for line in cmdbin.run_res_lines_stripped_searching(ros_res,'backports'):
                        print line
                    """ Rather than have the entire CMD_APT_UPDATE output
                    displayed on screen we print just those lines containing 'backports'
                    """

    if has_backports_flag is False:
        """ Best attempt at adding backports into sources.list.d did not work
        so nothing more we can do """
        return False

    requires_res = False

    with settings(warn_only=True):
        apt_line = "{0} install {1}".format(CMD_APT,packagename)
        if assumeyes:
            if recommends is False:
                if confold:
                    apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS_CONFOLD,
                                                        packagename)
                else:
                    apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS,
                                                        packagename)
            else:
                apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES,packagename)
        else:
            if recommends is False:
                apt_line = "{0} install {1}".format(CMD_APT_NORECOMMENDS,packagename)
        ros_res = cmdbin.run_or_sudo(apt_line)
        if ros_res.succeeded:
            requires_res = True  

    status_list = dpkg_status_lines_list(packagename)
    print "{0}:{1} - {2}".format(requires_res,packagename,status_list)
    if len(status_list) > 0:
        return True

    return requires_res


def requires_backported_packages(package_list=None,assumeyes=False,
                                 recommends=True,confold=False,exists_replace=False):
    if package_list:
        if hasattr(package_list, "__iter__"):
            if type(package_list) == type(''):
                package_list = [package_list]
        else:
            return False
    else:
        return False

    version_qualifier_count = 0
    version_qualifier_dict = {}
    # chr(61) is equals
    for pkg in package_list:
        if chr(61) in pkg:
            version_qualifier_count += 1
            pkg_array = pkg.split(chr(61))
            version_qualifier_dict[pkg_array[0]]=pkg_array[1]
    #print len(version_qualifier_dict),version_qualifier_dict

    if version_qualifier_count == 0:
        """ When we are not given any version qualified packages
        we can shortcut to return True if all the package names
        are marked installed already
        """
        #if len(dpkg_status_lines_list(packagename)) > 0:
        if dpkg_status_installed(package_list):
            return True
    elif exists_replace is True:
        pass
    else:
        """ Some of required packages have version qualifiers
        and we are not given authority to replace existing packages
        """
        for pkg,version in version_qualifier_dict.iteritems():
            #if dpkg_status_installed([pkg]):
            #print pkg,version
            version_list = dpkg_installed_version([pkg])
            if len(version_list) > 0:
                installed_version = version_list[0]
                if installed_version == version.strip():
                    pass
                else:
                    print "Package {0} already installed. Will not replace.".format(pkg)
                    return False

    requires_res = True  

    BACKPORTS='backports'

    has_backports_flag = False
    if has_backports():
        has_backports_flag = True
    else:
        (pname, pretty_name_id, pretty_name_word) = cmdbin.pretty_name_tuple()
        pretty_dict = {'id': pretty_name_id,
                       'release': pretty_name_word,
                       'repolabel': BACKPORTS}
        sl_filename = "{0}.list".format(BACKPORTS)
        sl_lines = source_list_lines_debian(pretty_dict,BACKPORTS)
        if sl_lines:
            put_flag = cmdbin.put_using_named_temp_sudo("/etc/apt/sources.list.d/{0}".format(sl_filename),
                                                   sl_lines)
            if put_flag:
                has_backports_flag = True
                list_dot_d_array_extend([sl_filename])
                #print get_list_dot_d_array()
                with hide('output'):
                    ros_res = cmdbin.run_or_sudo(CMD_APT_UPDATE)
                    for line in cmdbin.run_res_lines_stripped_searching(ros_res,'backports'):
                        print line
                    """ Rather than have the entire CMD_APT_UPDATE output
                    displayed on screen we print just those lines containing 'backports'
                    """

    if has_backports_flag is False:
        """ Best attempt at adding backports into sources.list.d did not work
        so nothing more we can do """
        return False

    packages_count = len(package_list)
    packages_string = ' '.join(package_list)
    """ Above we have joined and will rely on apt-get to handle the requested
    package list (including any version qualified packages) rather than
    doing any further checking against required versions.
    We make no further use of version_qualifier_dict dictionary, which has
    thus far, only been used for handling exists_replace=False correctly.
    """

    requires_res = False

    with settings(warn_only=True):
        apt_line = "{0} install {1}".format(CMD_APT,packages_string)
        if assumeyes:
            if recommends is False:
                if confold:
                    apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS_CONFOLD,
                                                        packages_string)
                else:
                    apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS,
                                                        packages_string)
            else:
                apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES,packages_string)
        else:
            if recommends is False:
                apt_line = "{0} install {1}".format(CMD_APT_NORECOMMENDS,packages_string)
        ros_res = cmdbin.run_or_sudo(apt_line)
        if ros_res.succeeded:
            requires_res = True  

    if requires_res and version_qualifier_count == 0:
        """ When we believe successful and are simple case (no version qualifiers)
        provide an extra print and outstanding check
        """
        outstanding_list = dpkg_status_installed(package_list,'outstanding')
        if len(outstanding_list) > 0:
            print "backports install not completed for packages: {0}".format(outstanding_list)
            return False

    return requires_res


def requires_binary_package(binary_unpathed=None,packagename_suggestion=None,
                            assumeyes=False):
    """ The word 'binary' here refers to the fact that you supply a binary
    name as the first argument. Two examples follow:
    requires_res = requires_binary_package('ab','apache2-utils',assumeyes=True)
    requires_binary_package('django-admin','python-django',assumeyes=True)
    """
    which_res = None
    requires_which = cmdbin.CMD_KERN_RELEASE
    with settings(warn_only=True):
        which_res = run("{0} {1}".format(cmdbin.CMD_WHICH,binary_unpathed))

    print "{0} {1}".format(which_res.succeeded,which_res.failed)
    if which_res and which_res.succeeded:
        """ .succeeded is supported in Fabric 1.0 and newer """
        requires_which = which_res
        print which_res
        return requires_which
    elif which_res is not None and which_res.failed:
        if assumeyes:
            cmdbin.run_or_sudo("{0} install {1}".format(
                    CMD_APT_ASSUMEYES,packagename_suggestion))
        else:
            cmdbin.run_or_sudo("{0} install {1}".format(
                    CMD_APT,packagename_suggestion))

        with settings(warn_only=True):
            which_res = run("{0} {1}".format(cmdbin.CMD_WHICH,binary_unpathed))

        if which_res and not which_res.failed:
            requires_which = which_res
            print which_res
            return requires_which
    else:
        return which_res
    return False


def netselect_apt(verbose=False):
    netselect_return = False
    requires_res = requires_packages(['netselect','netselect-apt'],True,False,False)
    if verbose:
        # hide warnings but show stdout
        with hide('warnings'):
            ros_res = cmdbin.run_or_sudo(CMD_NETSELECT_APT)
    else:
        # hide warnings and stdout
        ros_res = cmdbin.run_or_sudo_warningless(CMD_NETSELECT_APT)
    if ros_res.succeeded and ros_res.find('Writing sources.list'):
        netselect_return = True
        print 'A new sample sources.list has been generated (/root/sources.list?)'
        print 'Inspect and if necessary copy into /etc/apt/ the results you are happy with'
    return netselect_return




